"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.commandNames = exports.shorthands = void 0;
exports.rcOptionsTypes = rcOptionsTypes;
exports.cliOptionsTypes = cliOptionsTypes;
exports.help = help;
exports.handler = handler;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const pick_1 = __importDefault(require("ramda/src/pick"));
const cli_utils_1 = require("@pnpm/cli-utils");
const config_1 = require("@pnpm/config");
const directory_fetcher_1 = require("@pnpm/directory-fetcher");
const fs_indexed_pkg_importer_1 = require("@pnpm/fs.indexed-pkg-importer");
const fs_is_empty_dir_or_nothing_1 = require("@pnpm/fs.is-empty-dir-or-nothing");
const plugin_commands_installation_1 = require("@pnpm/plugin-commands-installation");
const common_cli_options_help_1 = require("@pnpm/common-cli-options-help");
const error_1 = require("@pnpm/error");
const lockfile_fs_1 = require("@pnpm/lockfile.fs");
const rimraf_1 = __importDefault(require("@zkochan/rimraf"));
const render_help_1 = __importDefault(require("render-help"));
const deployHook_1 = require("./deployHook");
const logger_1 = require("@pnpm/logger");
const createDeployFiles_1 = require("./createDeployFiles");
const deployCatalogHook_1 = require("./deployCatalogHook");
const FORCE_LEGACY_DEPLOY = 'force-legacy-deploy';
exports.shorthands = {
    ...plugin_commands_installation_1.install.shorthands,
    legacy: [`--config.${FORCE_LEGACY_DEPLOY}=true`],
};
const DEPLOY_OWN_OPTIONS = (0, pick_1.default)([FORCE_LEGACY_DEPLOY], config_1.types);
function rcOptionsTypes() {
    return {
        ...plugin_commands_installation_1.install.rcOptionsTypes(),
        ...DEPLOY_OWN_OPTIONS,
    };
}
function cliOptionsTypes() {
    return {
        ...plugin_commands_installation_1.install.cliOptionsTypes(),
        ...DEPLOY_OWN_OPTIONS,
    };
}
exports.commandNames = ['deploy'];
function help() {
    return (0, render_help_1.default)({
        description: 'Experimental! Deploy a package from a workspace',
        url: (0, cli_utils_1.docsUrl)('deploy'),
        usages: ['pnpm --filter=<deployed project name> deploy <target directory>'],
        descriptionLists: [
            {
                title: 'Options',
                list: [
                    {
                        description: "Packages in `devDependencies` won't be installed",
                        name: '--prod',
                        shortAlias: '-P',
                    },
                    {
                        description: 'Only `devDependencies` are installed',
                        name: '--dev',
                        shortAlias: '-D',
                    },
                    {
                        description: '`optionalDependencies` are not installed',
                        name: '--no-optional',
                    },
                    {
                        description: 'Force legacy deploy implementation',
                        name: '--legacy',
                    },
                ],
            },
            common_cli_options_help_1.FILTERING,
        ],
    });
}
async function handler(opts, params) {
    if (!opts.workspaceDir) {
        throw new error_1.PnpmError('CANNOT_DEPLOY', 'A deploy is only possible from inside a workspace');
    }
    const selectedProjects = Object.values(opts.selectedProjectsGraph ?? {});
    if (selectedProjects.length === 0) {
        throw new error_1.PnpmError('NOTHING_TO_DEPLOY', 'No project was selected for deployment');
    }
    if (selectedProjects.length > 1) {
        throw new error_1.PnpmError('CANNOT_DEPLOY_MANY', 'Cannot deploy more than 1 project');
    }
    if (params.length !== 1) {
        throw new error_1.PnpmError('INVALID_DEPLOY_TARGET', 'This command requires one parameter');
    }
    const selectedProject = selectedProjects[0].package;
    const deployDirParam = params[0];
    const deployDir = path_1.default.isAbsolute(deployDirParam) ? deployDirParam : path_1.default.join(opts.dir, deployDirParam);
    if (!(0, fs_is_empty_dir_or_nothing_1.isEmptyDirOrNothing)(deployDir)) {
        if (!opts.force) {
            throw new error_1.PnpmError('DEPLOY_DIR_NOT_EMPTY', `Deploy path ${deployDir} is not empty`);
        }
        logger_1.logger.warn({ message: 'using --force, deleting deploy path', prefix: deployDir });
    }
    await (0, rimraf_1.default)(deployDir);
    await fs_1.default.promises.mkdir(deployDir, { recursive: true });
    const includeOnlyPackageFiles = !opts.deployAllFiles;
    await copyProject(selectedProject.rootDir, deployDir, { includeOnlyPackageFiles });
    if (opts.sharedWorkspaceLockfile) {
        const warning = opts.forceLegacyDeploy
            ? 'Shared workspace lockfile detected but configuration forces legacy deploy implementation.'
            : await deployFromSharedLockfile(opts, selectedProject, deployDir);
        if (warning) {
            (0, logger_1.globalWarn)(warning);
        }
        else {
            return;
        }
    }
    const deployedProject = opts.allProjects?.find(({ rootDir }) => rootDir === selectedProject.rootDir);
    if (deployedProject) {
        deployedProject.modulesDir = path_1.default.relative(selectedProject.rootDir, path_1.default.join(deployDir, 'node_modules'));
    }
    await plugin_commands_installation_1.install.handler({
        ...opts,
        confirmModulesPurge: false,
        // Deploy doesn't work with dedupePeerDependents=true currently as for deploy
        // we need to select a single project for install, while dedupePeerDependents
        // doesn't work with filters right now.
        // Related issue: https://github.com/pnpm/pnpm/issues/6858
        dedupePeerDependents: false,
        // If enabled, dedupe-injected-deps will symlink workspace packages in the
        // deployed dir to their original (non-deployed) directory in an attempt to
        // dedupe workspace packages that don't need to be injected. The deployed
        // dir shouldn't have symlinks to the original workspace. Disable
        // dedupe-injected-deps to always inject workspace packages since copying is
        // desirable.
        dedupeInjectedDeps: false,
        // Compute the wanted lockfile correctly by setting pruneLockfileImporters.
        // Since pnpm deploy only installs dependencies for a single selected
        // project, other projects in the "importers" lockfile section will be
        // empty when node-linker=hoisted.
        //
        // For example, when deploying project-1, project-2 may not be populated,
        // even if it has dependencies.
        //
        //   importers:
        //     project-1:
        //       dependencies:
        //         foo:
        //           specifier: ^1.0.0
        //           version: ^1.0.0
        //     project-2: {}
        //
        // Avoid including these empty importers in the in-memory wanted lockfile.
        // This is important when node-linker=hoisted to prevent project-2 from
        // being included in the hoisted install. If project-2 is errantly hoisted
        // to the root node_modules dir, downstream logic will fail to inject it to
        // the deploy directory. It's also just weird to include empty importers
        // that don't matter to the filtered lockfile generated for pnpm deploy.
        pruneLockfileImporters: true,
        depth: Infinity,
        hooks: {
            ...opts.hooks,
            readPackage: [
                ...(opts.hooks?.readPackage ?? []),
                deployHook_1.deployHook,
                deployCatalogHook_1.deployCatalogHook.bind(null, opts.catalogs ?? {}),
            ],
        },
        frozenLockfile: false,
        preferFrozenLockfile: false,
        // Deploy doesn't work currently with hoisted node_modules.
        // TODO: make it work as we need to prefer packages from the lockfile during deployment.
        useLockfile: opts.nodeLinker !== 'hoisted',
        saveLockfile: false,
        virtualStoreDir: path_1.default.join(deployDir, 'node_modules/.pnpm'),
        modulesDir: path_1.default.relative(opts.workspaceDir, path_1.default.join(deployDir, 'node_modules')),
        rawLocalConfig: {
            ...opts.rawLocalConfig,
            // This is a workaround to prevent frozen install in CI envs.
            'frozen-lockfile': false,
        },
        includeOnlyPackageFiles,
    });
}
async function copyProject(src, dest, opts) {
    const { filesIndex } = await (0, directory_fetcher_1.fetchFromDir)(src, opts);
    const importPkg = (0, fs_indexed_pkg_importer_1.createIndexedPkgImporter)('clone-or-copy');
    importPkg(dest, { filesMap: filesIndex, force: true, resolvedFrom: 'local-dir' });
}
async function deployFromSharedLockfile(opts, selectedProject, deployDir) {
    if (!opts.injectWorkspacePackages) {
        throw new error_1.PnpmError('DEPLOY_NONINJECTED_WORKSPACE', 'By default, starting from pnpm v10, we only deploy from workspaces that have "inject-workspace-packages=true" set', {
            hint: 'If you want to deploy without using injected dependencies, run "pnpm deploy" with the "--legacy" flag or set "force-legacy-deploy" to true',
        });
    }
    const { allProjects, lockfileDir, rootProjectManifest, rootProjectManifestDir, workspaceDir, } = opts;
    // The following errors should not be possible. It is a programmer error if they are reached.
    if (!allProjects)
        throw new Error('opts.allProjects is undefined.');
    if (!lockfileDir)
        throw new Error('opts.lockfileDir is undefined.');
    if (!workspaceDir)
        throw new Error('opts.workspaceDir is undefined.');
    const lockfile = await (0, lockfile_fs_1.readWantedLockfile)(lockfileDir, { ignoreIncompatible: false });
    if (!lockfile) {
        return 'Shared lockfile not found. Falling back to installing without a lockfile.';
    }
    const projectId = (0, lockfile_fs_1.getLockfileImporterId)(lockfileDir, selectedProject.rootDir);
    const deployFiles = (0, createDeployFiles_1.createDeployFiles)({
        allProjects,
        deployDir,
        lockfile,
        lockfileDir,
        rootProjectManifest,
        selectedProjectManifest: selectedProject.manifest,
        projectId,
        rootProjectManifestDir,
    });
    await Promise.all([
        fs_1.default.promises.writeFile(path_1.default.join(deployDir, 'package.json'), JSON.stringify(deployFiles.manifest, undefined, 2) + '\n'),
        (0, lockfile_fs_1.writeWantedLockfile)(deployDir, deployFiles.lockfile),
    ]);
    try {
        await plugin_commands_installation_1.install.handler({
            ...opts,
            allProjects: undefined,
            allProjectsGraph: undefined,
            selectedProjectsGraph: undefined,
            rootProjectManifest: deployFiles.manifest,
            rootProjectManifestDir: deployDir,
            dir: deployDir,
            lockfileDir: deployDir,
            workspaceDir: undefined,
            virtualStoreDir: undefined,
            modulesDir: undefined,
            confirmModulesPurge: false,
            frozenLockfile: true,
            hooks: {
                ...opts.hooks,
                readPackage: [
                    ...(opts.hooks?.readPackage ?? []),
                    deployHook_1.deployHook,
                    deployCatalogHook_1.deployCatalogHook.bind(null, opts.catalogs ?? {}),
                ],
                calculatePnpmfileChecksum: undefined, // the effects of the pnpmfile should already be part of the package snapshots
            },
            rawLocalConfig: {
                ...opts.rawLocalConfig,
                'frozen-lockfile': true,
            },
        });
    }
    catch (error) {
        (0, logger_1.globalWarn)('Deployment with a shared lockfile has failed. If this is a bug, please report it at <https://github.com/pnpm/pnpm/issues>.');
        (0, logger_1.globalWarn)(`As a workaround, you may add ${FORCE_LEGACY_DEPLOY}=true to .npmrc.`);
        throw error;
    }
    return undefined;
}
//# sourceMappingURL=deploy.js.map